;;; proj9.el --- Project Management

;;  (c) copyright 2001,2003,2007 Stefan Reuther

;; Author: Stefan Reuther <Streu@gmx.de>
;; X-URL: http://phost.de/~stefan/emacs/

;; This program is free software; you can redistribute it and/or
;; modify it under the terms of the GNU General Public License as
;; published by the Free Software Foundation; either version 2, or (at
;; your option) any later version.

;; This program is distributed in the hope that it will be useful, but
;; WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;; General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with [your] Emacs; see the file COPYING.  If not, write to the
;; Free Software Foundation, Inc., 59 Temple Place - Suite 330,
;; Boston, MA 02111-1307, USA.

;;; Commentary:

;; This is a *very* simple project manager. It consists of this
;; .el file and an associated `proj3.pl'/`proj4.pl'/`proj6.pl' Perl script.
;; Its main task is automatic creation of Makefiles.

(defvar project-root nil)
(defvar project-manager-v3 "proj4")
(defvar project-manager-v6 "proj6")
(defvar project-manager-v9 "proj9")
(defvar project-manager project-manager-v3)
(defvar project-grep-command "grep -n")

(defvar project-mode-map
  nil
  "Keymap for project editor")

(defun project-relative-file-name (FNAME)
  "Return FNAME as project-relative file name

If FNAME is below `project-root' in the directory tree,
returns the relative path; nil otherwise."
  (let ((ROOT (expand-file-name project-root)))
    (if (and (< (length ROOT) (length FNAME))
	     (equal ROOT (substring FNAME 0 (length ROOT))))
	(let ((FN (substring FNAME (length ROOT))))
	  (if (equal (substring FN 0 1) "/")
	      (setq FN (substring FN 1)))
	  FN)
      nil))
  )

(defun project-submit-command (CMD)
  "Process specified command in project manager

Invokes the project manager (`project-manager') on the
current project (`project-root'), with the specified
command (e.g. \"add-file foo.c\"). Returns the process
manager's output. \"no operation\" messages (informative)
are not returned."
  (let ((CMD (concat project-manager " root=" project-root " quiet=yes " CMD))
	MSG)
    ;(message CMD)
    (setq MSG (shell-command-to-string CMD))
    (if (string-match "no operation" MSG)
	""
      MSG)
    )
  )

(defun project-submit-command-echo (CMD)
  "Process specified command in project manager

Like `project-submit-command', but echoes the project manager
diagnostics in the echo area, if some are issued."
  (let ((RET (project-submit-command CMD)))
    (unless (equal RET "")
      (message "%s" RET)))
  ;; FIXME: put it in some buffer and display it, if needed
  )

(defun project-remove-this-file ()
  "Remove current file from project

Removes the current file from the project, if possible.
If the current file does not belong to the project, does
nothing."
  (interactive)
  (let (FN RET)
    (if (and project-root
	     (setq FN (project-relative-file-name (buffer-file-name))))
	(progn
	  (setq RET (project-submit-command (concat " remove-file " FN)))
	  (if (not (equal RET ""))
	      (message "%s" RET))
	  )
      )
    )
  )

(defun project-maybe-add-file ()
  (let (FN RET CMD)
    (if (and project-root
	     (setq FN (project-relative-file-name (buffer-file-name))))
	(project-submit-command-echo (concat " add-file " FN)))
    )
  )

(defun project-add-this-file ()
  "Add file to project

Adds the current file to the project. Does nothing if
already added or adding not possible (file outside
project tree)."
  (interactive)
  (project-maybe-add-file)
  )

(defun project-desktop-after-load ()
  "Project/Desktop Association

Add this to `desktop-after-load-hook' to have your
project automatically associated with the desktop(s)
in the same directory. You have to do a \"proj3 init\"
manually before switching to that desktop to set this
up."
  (if (and (boundp 'desktop-file) desktop-file)
      (let ((DIR (expand-file-name (file-name-directory desktop-file))))
	(if (string-match "/$" DIR)
	    (setq DIR (substring DIR 0 -1)))
	(if (file-readable-p (concat DIR "/P9/Settings"))
	    (setq project-root DIR
                  project-manager project-manager-v9)
          (if (file-readable-p (concat DIR "/P/DEFS"))
              (setq project-root DIR
                    project-manager project-manager-v3)
            (if (file-readable-p (concat DIR "/P/Settings"))
                (setq project-root DIR
                      project-manager project-manager-v6)
              (setq project-root nil))))
	)
    )
  )

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; Commands ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun project-make ()
  "Make project

When you're editing a project, invoke \"make\" on it (by
calling `compile' with \"make -k -C `project-root'\").
Otherwise, behaves exactly like `compile'."
  (interactive)
  (unless (project-maybe-insert)
    (if (and project-root (not current-prefix-arg))
	(let (compile-command)
	  (compile (concat "make -k -C " project-root)))
      (call-interactively 'compile))
    )
  )

;"
(defun project-grep (COMMAND)
  "Grep in project

Like \\[grep], but when editing a project, defaults
to searching in the project's files only.

When called interactively, prompts for string to grep for,
or, when called with prefix argument, behaves 100% like
\\[grep]"
  (interactive (if current-prefix-arg
		   (list (read-shell-command 
			  "Run grep (like this): "
			  grep-command 'grep-history))
		 (list (concat project-grep-command " "
			       (read-string "Grep for: ")))))
  (if project-root
      (let ((default-directory (concat project-root "/")))
	(grep (concat project-manager " root=" project-root " list files | xargs " COMMAND)))
    (grep COMMAND))
)

(defun project-update ()
  (interactive)
  (message "Updating project...")
  (unless
      (project-submit-command-echo " update")
    (message "Updating project...done"))
  )

(defun project-maybe-update ()
  (let (fn pfx)
    (if (and project-root
	     (setq fn (buffer-file-name))
	     (or (and (setq pfx (concat project-root "/P/"))
                      (> (length fn) (length pfx))
                      (equal (substring fn 0 (length pfx)) pfx))
                 (and (setq pfx (concat project-root "/P9/"))
                      (> (length fn) (length pfx))
                      (equal (substring fn 0 (length pfx)) pfx))))
        (project-update))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; Interactive Project Editor

(defun project-edit ()
  "Edit project interactively"
  (interactive)
  (if (not project-root)
      (error "Not editing a project")
    (let ((BUFFER (get-buffer-create "*Project Editor*")))
      (switch-to-buffer BUFFER)
      (project-get-project)
      (project-edit-mode)
      ))
  )

(define-error 'proj-invalid-buffer "Attempt to call project-get-project in invalid buffer" 'invalid-operation)

; Internal: show project manager in current buffer
(defun project-get-project ()
  (interactive)
  (if (buffer-file-name)
      (error 'proj-invalid-buffer))
  (let ((inhibit-read-only t))
    (delete-region (point-min) (point-max)))
  (save-excursion
    ;(message (prin1-to-string (list project-manager nil (current-buffer) nil (concat "root=" project-root) "show" "default"
    ;                                )))
    (call-process project-manager nil (current-buffer) nil (concat "root=" project-root) "show" "default"))
  ;; Assign text properties
  (let (END NAME)
    (goto-char (point-max))
    (setq END (point))
    (while (search-backward-regexp "^==\\s +\\([^=]+\\b\\)\\s +=" nil t)
      (setq NAME (match-string 1))
      ;; Remember name of this section
      (put-text-property (point) END 'project-section-name NAME)
      ;; Mark section contents read-only. In the `Build Rules'
      ;; section, only the actual commands are editable.
      (if (string= NAME "Build Rules")
	  (save-excursion
	    (forward-line 2)
	    (setq END (point))))
      (put-text-property (point) END 'read-only t)
      (setq END (point))
      ))
  )

(defun project-maybe-insert ()
  (interactive)
  (if (and last-command-event
	   (characterp last-command-event)
	   (not current-prefix-arg)
	   (not (get-text-property (point) 'read-only)))
      (progn 
	(insert (if (equal last-command-event ?\r) "\n" last-command-event))
	t)
    nil)
  )

(if project-mode-map
    nil
  (setq project-mode-map (make-keymap))
  ;(suppress-keymap project-mode-map)
  (define-key project-mode-map "x" 'project-mode-submit)
  (define-key project-mode-map "g" 'project-mode-reload)
  (define-key project-mode-map " " 'project-mode-toggle)
  (define-key project-mode-map "." 'project-mode-toggle)
  (define-key project-mode-map "d" 'project-mode-delete)
  (define-key project-mode-map "u" 'project-mode-undelete)
  (define-key project-mode-map "o" 'project-mode-open)
  (define-key project-mode-map "e" 'project-mode-open)
  (define-key project-mode-map [(ret)] 'project-mode-open)
  (define-key project-mode-map [(newline)] 'project-mode-open)
  (define-key project-mode-map "\C-m" 'project-mode-open)
  (define-key project-mode-map "\n" 'project-mode-open)
  (define-key project-mode-map "w" 'project-mode-who-uses)
  (define-key project-mode-map "m" 'project-make)
  (define-key project-mode-map "a" 'project-mode-add)
  )

(defun project-mode-add ()
  (interactive)
  (unless (project-maybe-insert)
    (let ((section (get-text-property (point) 'project-section-name)))
      (if (or (string= section "Project Files")
	      (string= section "File / Project Matrix"))
	  (call-interactively 'project-mode-add-internal)))
    ))

(defun project-mode-add-internal (FILE)
  ;; FIXME: not yet implemented
  (interactive "FAdd file to project: ")
  (setq FILE (project-relative-file-name (expand-file-name FILE)))
  (if (not FILE)
      (error "File is not in project tree"))
  ;(project-submit-command-echo (concat "blacklist=no add-file " FILE))
  ; NO! just add to display; will be processed later when user hits `x'
  )

(defun project-mode-who-uses ()
  "Look who uses a particular file

A wrapper around \\[project-grep]. When called interactively,
looks for uses of the file the cursor is on; when called with
prefix argument, prompts for a file name."
  (interactive)
  (unless (project-maybe-insert)
    (call-interactively 'project-mode-who-uses-internal)))

(defun project-mode-who-uses-internal (FILE)
  (interactive
   (list (if current-prefix-arg
	     (read-string "Who uses file: ")
	   (save-excursion
	     (beginning-of-line)
	     (and (looking-at "\\([^ ]*\\) ([^)]*)")
		  (match-string 1))))))
  (if (and FILE (not (equal "" FILE)))
      (project-grep (concat project-grep-command " '[\"/]"
			    (regexp-quote FILE) "\"'"))
    (message "No filename specified")
    )
  )

(defun project-mode-open ()
  "Open current file

In the project manager, open the file the cursor is at (by
means of `find-file')."
  (interactive)
  (unless (project-maybe-insert)
    (let (POS NAME)
      (save-excursion
	(beginning-of-line)
	(cond
	 ((looking-at "\\(  \\|\\)\\([^ \n]*\\) ([^)]*)")
	  (find-file (match-string 2)))
	 ((looking-at "- Depends on \\([^ \n]+\\)")
	  (setq NAME (match-string 1))
	  (and (search-backward-regexp "^==" nil t)
	       (or (forward-line 1) t)
	       (search-forward-regexp (concat "^\\(Project " NAME " *(\\|==\\)") nil t)
	       (or (beginning-of-line) t)
	       (looking-at "^Project")
	       (setq POS (point))))
	 t))
      (if POS
	  (goto-char POS))
      )
    )
  )

(defun project-mode-toggle ()
  "Toggle project membership of a file

Use to toggle if a file is member of a particular subproject
or not."
  (interactive)
  (unless (project-maybe-insert)
    (let ((inhibit-read-only t))
      (while (not (or (bolp)
		      (looking-at "\\[.\\]")))
	(forward-char -1))
      (cond
       ((looking-at "\\[[xX]\\]")
	(delete-char 3)
      (insert "[ ]")
      (forward-char -2))
       ((looking-at "\\[.\\]")
	(delete-char 3)
	(insert "[x]")
	(forward-char -2)))
      )
    )
  )

(defun project-mode-delete ()
  "Mark file for deletion from project

Mark file with \"(Deleted)\" in the last column. It will be
removed from the project the next time you use \\[project-mode-submit].
The actual disk file will be unaffected by this."
  (interactive)
  (unless (project-maybe-insert)
    (save-excursion
      (let ((inhibit-read-only t))
	(beginning-of-line)
	(if (or (looking-at "^[^\n]*\\[.\\]")	; line with checkbox
		(looking-at "^Project [^\n](")
		(looking-at "^- Depends on ")
		(looking-at "^  "))
	    (if (looking-at "^[^\n]*(Deleted)")
		nil
	    (end-of-line)
	    (if (< (current-column) 50)
		(insert (repeat-string (- 50 (current-column)) " ")))
	    (insert "  (Deleted)"))
	  )
	)
      )
    )
  )

(defun project-mode-undelete ()
  "Mark file undeleted

Remove the deletion marker (\\[project-mode-delete]) from
the current file."
  (interactive)
  (unless (project-maybe-insert)
    (save-excursion
      (let ((inhibit-read-only t))
	(beginning-of-line)
	(if (or (looking-at "^[^\n]*\\[.\\]")	; line with checkbox
		(looking-at "^Project [^\n](")
		(looking-at "^- Depends on ")
		(looking-at "^  "))
	    (if (looking-at "^[^\n]*(Deleted)")
		(progn
		  (search-forward-regexp " *(Deleted)")
		  (delete-region (match-beginning 0) (match-end 0)))
	      )
	  )
	)
      )
    )
  )

(defun project-mode-execute ()
  "Execute changes to project manager

This will process subproject membership changes and delete markers.
This will *not* update display for other changes, use
\\[project-mode-submit] for that."
  (interactive)
  (unless (project-maybe-insert)
    (let ((log-buffer (get-buffer-create " *Project Manager Output*"))
	  origpt)
      (delete-region (point-min log-buffer) (point-max log-buffer) log-buffer)
      (setq origpt (point log-buffer))
      (call-process-region (point-min) (point-max)
			   project-manager
			   nil log-buffer nil 
			   (concat "root=" project-root) "read")
      (if (not (equal origpt (point log-buffer)))
	  (progn
	    (message "%s" (buffer-substring origpt (point log-buffer) log-buffer))
	    nil)
	(message "Project updated.")
	t
	)
      )
    )
  )

(defun project-mode-submit ()
  "Submit changes to project manager, and reload"
  (interactive)
  (unless (project-maybe-insert)
    (if (project-mode-execute)
	(project-get-project))
    )
  )

(defun project-mode-reload ()
  "Reload project manager. This will discard all changes, and
load external modifications."
  (interactive)
  (unless (project-maybe-insert)
    (project-get-project)))

(defun project-begin (DIR)
  "Start editing a project"
  (interactive "DProject Root Directory: ")
  (unless (and (stringp DIR)
	       (file-directory-p DIR)
	       (not (equalp "" DIR)))
    (error "This is not an existing directory."))
  (setq DIR (expand-file-name DIR))
  (if (equalp (substring DIR -1) "/")
      (setq DIR (substring DIR 0 -1)))

  (unless (or (file-exists-p (concat DIR "/P/DEFS"))
              (file-exists-p (concat DIR "/P/Settings"))
              (file-exists-p (concat DIR "/P9/Settings")))
    (let ((project-root DIR))
      (project-submit-command-echo (concat "init "
					   (read-string "Name of main project: ")))))
  (setq project-root DIR)
  )

(defun project-edit-mode ()
  "Project Editor Mode

Mode for editing projects. Use \\[project-edit] to load this mode.
Keys are similar (sort-of) to `dired' and `buffer-menu'.

Commands are normal letter keys. In some sections these insert
themselves; use C-u <key> to get the command then.

\\{project-mode-map}"
  (interactive)
  (kill-all-local-variables)
  (use-local-map project-mode-map)
  (setq major-mode 'project-edit-mode
	mode-name "Project Editor")
  (setq font-lock-keywords
	'(("\\[.\\]" 0 bold)
	  ("^\\(  \\|\\)\\([^ ]*\\) ([^)]*)" (2 font-lock-reference-face t))
	  ("^==[^\n]+" 0 standout)
	  ("^Project +[^\n (]+ *([^) \n]+)" 0 bold)
	  ("^- Depends on \\([^ \n]*\\)" (1 font-lock-reference-face t))
	  ("([Dd]eleted?)$" 0 italic)
	  ("^Project.*\\(([Cc]urrent?)\\)" 1 italic)
	  ))
  (run-hooks 'project-edit-mode-hook)
  )

(add-hook 'find-file-not-found-hooks 'project-maybe-add-file)
(add-hook 'after-save-hook 'project-maybe-update)
